<?php

namespace Drupal\commerce_iats\Plugin\Commerce\PaymentGateway;

use CommerceGuys\Intl\Formatter\NumberFormatterInterface;
use Drupal\commerce_iats\Exception\GatewayException;
use Drupal\commerce_iats\Exception\GenericPaymentGatewayException;
use Drupal\commerce_payment\CreditCard;
use Drupal\commerce_payment\Entity\PaymentInterface;
use Drupal\commerce_payment\Entity\PaymentMethodInterface;
use Drupal\commerce_payment\Exception\HardDeclineException;
use Drupal\commerce_payment\Exception\PaymentGatewayException;
use Drupal\commerce_payment\Plugin\Commerce\PaymentGateway\OnsitePaymentGatewayBase;
use Drupal\commerce_payment\Plugin\Commerce\PaymentGateway\SupportsAuthorizationsInterface;
use Drupal\commerce_payment\Plugin\Commerce\PaymentGateway\SupportsRefundsInterface;
use Drupal\commerce_payment\Plugin\Commerce\PaymentGateway\SupportsUpdatingStoredPaymentMethodsInterface;
use Drupal\commerce_price\Price;
use Drupal\Core\Form\FormStateInterface;

/**
 * Class PaymentForm.
 *
 * @CommercePaymentGateway(
 *   id = "commerce_iats",
 *   label = "iATS",
 *   display_label = "iATS",
 *   forms = {
 *     "add-payment-method" = "Drupal\commerce_iats\PluginForm\PaymentMethodAddForm",
 *     "capture-payment" = "Drupal\commerce_payment\PluginForm\PaymentCaptureForm",
 *   },
 *   payment_method_types = {"credit_card"},
 *   credit_card_types = {
 *     "amex", "discover", "mastercard", "visa",
 *   },
 *   modes = {"live" = "Live"},
 *   js_library = "commerce_iats/cryptogram",
 * )
 */
class CommerceIatsGateway extends OnsitePaymentGatewayBase implements SupportsUpdatingStoredPaymentMethodsInterface, SupportsAuthorizationsInterface, SupportsRefundsInterface {

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration() {
    return [
      'transcenter' => '',
      'processor' => '',
      'gateway_id' => '',
      'processing_type' => 'hosted',
    ] + parent::defaultConfiguration();
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state) {
    $form = parent::buildConfigurationForm($form, $form_state);

    $form['transcenter'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Transaction Center ID'),
      '#description' => $this->t('Your Transaction Center ID number for your 1stPayGateway account.'),
      '#default_value' => $this->configuration['transcenter'],
      '#required' => TRUE,
    ];

    $form['processor'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Processor ID'),
      '#description' => $this->t('The ID number for a MID/TID combination on your 1stPayGateway account.'),
      '#default_value' => $this->configuration['processor'],
      '#required' => TRUE,
    ];

    $form['gateway_id'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Gateway ID'),
      '#description' => $this->t('The Gateway ID for your 1stPayGateway account.'),
      '#default_value' => $this->configuration['gateway_id'],
      '#required' => TRUE,
    ];

    $form['processing_type'] = [
      '#type' => 'select',
      '#title' => $this->t('Processing type'),
      '#description' => $this->t('The processing type to utilize for this gateway.'),
      '#default_value' => $this->configuration['processing_type'],
      '#required' => TRUE,
      '#options' => [
        'hosted' => $this->t('Hosted form'),
        'direct_submission' => $this->t('Direct submission'),
      ],
    ];

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function submitConfigurationForm(array &$form, FormStateInterface $form_state) {
    parent::submitConfigurationForm($form, $form_state);

    if (!$form_state->getErrors()) {
      $values = $form_state->getValue($form['#parents']);
      $this->configuration['transcenter'] = $values['transcenter'];
      $this->configuration['processor'] = $values['processor'];
      $this->configuration['gateway_id'] = $values['gateway_id'];
      $this->configuration['processing_type'] = $values['processing_type'];
    }
  }

  /**
   * {@inheritdoc}
   */
  public function createPaymentMethod(PaymentMethodInterface $payment_method, array $payment_details) {
    // Submit to the gateway to create a credit card record.
    $gateway = $this->getGateway();
    $vaultKey = $this->getCommerceIats()
      ->getUserVaultId($payment_method->getOwner());
    /** @var \Drupal\address\AddressInterface $address */
    $address = $payment_method->getBillingProfile()->address->first();

    $data = [
      'ownerName' => $address->getGivenName() . ' ' . $address->getFamilyName(),
      'ownerStreet' => $address->getAddressLine1(),
      'ownerStreet2' => $address->getAddressLine2(),
      'ownerCity' => $address->getLocality(),
      'ownerState' => $address->getAdministrativeArea(),
      'ownerZip' => $address->getPostalCode(),
      'ownerCountry' => $address->getCountryCode(),
      'ownerEmail' => $payment_method->getOwner()->getEmail(),
    ];

    if ($this->isHosted()) {
      $data['creditCardCryptogram'] = $payment_details['cryptogram'];
    }
    else {
      $data['cardNumber'] = $payment_details['number'];
      $data['cardExpMonth'] = $payment_details['expiration']['month'];
      $data['cardExpYear'] = substr($payment_details['expiration']['year'], -2);
      $data['cardType'] = $this->mapCreditCardType($payment_details['type'], TRUE);
    }

    // Add credit card to the vault.
    try {
      $result = $gateway->vaultCcCreate($vaultKey, $data);
      $id = $result->id;
    }
    catch (GatewayException $e) {
      $data = $e->getData()->data;
      if (empty($data->id)) {
        throw new GenericPaymentGatewayException();
      }
      $id = $data->id;
    }
    catch (\Exception $e) {
      throw new GenericPaymentGatewayException();
    }

    // Get the credit card details from the vault.
    try {
      $result = $gateway->vaultCcLoad($vaultKey, $id);
    }
    catch (\Exception $e) {
      throw new GenericPaymentGatewayException();
    }

    $payment_method->card_type = $this->mapCreditCardType($result->cardType);
    $payment_method->card_number = $result->cardNoLast4;
    $payment_method->card_exp_month = $result->cardExpMM;
    $payment_method->card_exp_year = $result->cardExpYY;
    $payment_method->setRemoteId($id);
    $expires = CreditCard::calculateExpirationTimestamp(
      $payment_method->card_exp_month->value,
      $payment_method->card_exp_year->value);
    $payment_method->setExpiresTime($expires);
    $payment_method->save();
  }

  /**
   * {@inheritdoc}
   */
  public function deletePaymentMethod(PaymentMethodInterface $payment_method) {
    $gateway = $this->getGateway();
    $vaultKey = 'drupal8_iats_commerce!' . \Drupal::currentUser()->id();
    $gateway->vaultCcDelete($vaultKey, $payment_method->getRemoteId());
    $payment_method->delete();
  }

  /**
   * {@inheritdoc}
   */
  public function createPayment(PaymentInterface $payment, $capture = TRUE) {
    $this->assertPaymentState($payment, ['new']);
    $payment_method = $payment->getPaymentMethod();
    $this->assertPaymentMethod($payment_method);
    $amount = $payment->getAmount();

    $account = $payment->getOrder()->getCustomer();
    $vaultKey = $this->getCommerceIats()->getUserVaultId($account);

    $transaction_data = [
      'orderId' => $payment->getOrderId() . '-' . $this->time->getCurrentTime(),
      'transactionAmount' => $this->formatAmount($amount->getNumber()),
    ];

    try {
      if ($capture) {
        $result = $this->getGateway()->firstPayCcSale(
          $vaultKey,
          $payment_method->getRemoteId(),
          $transaction_data
        );
        $payment->setState('completed');
      }
      else {
        $result = $this->getGateway()->firstPayCcAuth(
          $vaultKey,
          $payment_method->getRemoteId(),
          $transaction_data
        );
        $payment->setState('authorization');
        $payment->setAuthorizedTime(time());
        $payment->setExpiresTime(strtotime('now +30 days'));
      }
    }
    catch (\Exception $e) {
      throw new GenericPaymentGatewayException();
    }

    $payment->setRemoteId($result->referenceNumber);
    $payment->save();
  }

  /**
   * Maps iATS credit card types to Commerce credit card types.
   *
   * @param string $cardType
   *   The iATS credit card type.
   * @param bool $rev
   *   Reverse the lookup, to get iATS credit card type from Commerce credit
   *   card type.
   *
   * @return string
   *   The Commerce credit card type.
   */
  public function mapCreditCardType($cardType, $rev = FALSE) {
    $map = [
      'Amex' => 'amex',
      'Discover' => 'discover',
      'MasterCard' => 'mastercard',
      'Visa' => 'visa',
    ];

    if ($rev) {
      $map = array_flip($map);
    }

    if (!isset($map[$cardType])) {
      throw new HardDeclineException(sprintf('Unsupported credit card type "%s".', $cardType));
    }

    return $map[$cardType];
  }

  /**
   * {@inheritdoc}
   */
  public function capturePayment(PaymentInterface $payment, Price $amount = NULL) {
    $this->assertPaymentState($payment, ['authorization']);
    // If not specified, capture the entire amount.
    $amount = $amount ?: $payment->getAmount();

    $gateway = $this->getGateway();
    $data = [
      'refNumber' => $payment->getRemoteId(),
      'transactionAmount' => $this->formatAmount($amount->getNumber()),
    ];

    try {
      $gateway->creditCardSettle($data);
    }
    catch (\Exception $e) {
      throw new PaymentGatewayException($this->t('Unable to perform settlement.'));
    }

    $payment->setState('completed');
    $payment->setAmount($amount);
    $payment->save();
  }

  /**
   * {@inheritdoc}
   */
  public function voidPayment(PaymentInterface $payment) {
    $this->assertPaymentState($payment, ['authorization']);

    $gateway = $this->getGateway();
    $data = ['refNumber' => $payment->getRemoteId()];

    try {
      $gateway->creditCardVoid($data);
    }
    catch (\Exception $e) {
      throw new PaymentGatewayException($this->t('Unable to perform void.'));
    }

    $payment->setState('authorization_voided');
    $payment->save();
  }

  /**
   * {@inheritdoc}
   */
  public function updatePaymentMethod(PaymentMethodInterface $payment_method) {
    // TODO: Implement updatePaymentMethod() method.
  }

  /**
   * {@inheritdoc}
   */
  public function refundPayment(PaymentInterface $payment, Price $amount = NULL) {
    $this->assertPaymentState($payment, ['completed', 'partially_refunded']);
    // If not specified, refund the entire amount.
    $amount = $amount ?: $payment->getAmount();
    $this->assertRefundAmount($payment, $amount);

    $gateway = $this->getGateway();
    $data = [
      'refNumber' => $payment->getRemoteId(),
      'transactionAmount' => $this->formatAmount($amount->getNumber()),
    ];

    try {
      $gateway->creditCardCredit($data);
    }
    catch (\Exception $e) {
      throw new PaymentGatewayException($this->t('Unable to perform refund.'));
    }

    $old_refunded_amount = $payment->getRefundedAmount();
    $new_refunded_amount = $old_refunded_amount->add($amount);
    if ($new_refunded_amount->lessThan($payment->getAmount())) {
      $payment->setState('partially_refunded');
    }
    else {
      $payment->setState('refunded');
    }

    $payment->setRefundedAmount($new_refunded_amount);
    $payment->save();
  }

  /**
   * Determines if the payment gateway is operating with hosted form processing.
   *
   * @return bool
   *   Indicates if the payment gateway is operating with hosted form
   *   processing.
   */
  public function isHosted() {
    return $this->configuration['processing_type'] == 'hosted';
  }

  /**
   * Formats an amount.
   *
   * @param string $amount
   *   The raw amount.
   *
   * @return string
   *   The formatted amount.
   */
  protected function formatAmount($amount) {
    return $this->getAmountFormatter()->format($amount);
  }

  /**
   * Gets a number formatter ready to handle amounts.
   *
   * @return \CommerceGuys\Intl\Formatter\NumberFormatterInterface
   *   The number formatter.
   */
  protected function getAmountFormatter() {
    /** @var \Drupal\commerce_price\NumberFormatterFactoryInterface $formatterFactory */
    $formatterFactory = \Drupal::service('commerce_price.number_formatter_factory');
    $numberFormatter = $formatterFactory->createInstance(NumberFormatterInterface::DECIMAL);
    return $numberFormatter->setGroupingUsed(FALSE)
      ->setMinimumFractionDigits(2)
      ->setMaximumFractionDigits(2);
  }

  /**
   * Gets the commerce iATS rest gateway service.
   *
   * @return \Drupal\commerce_iats\Rest\GatewayInterface
   *   The commerce iATS rest gateway service.
   */
  protected function getGateway() {
    return \Drupal::service('commerce_iats.rest_gateway_factory')
      ->getGateway(
        $this->configuration['gateway_id'],
        $this->configuration['processor']
      );
  }

  /**
   * Gets the commerce iATS service.
   *
   * @return \Drupal\commerce_iats\CommerceIats
   *   The commerce iATS service.
   */
  protected function getCommerceIats() {
    return \Drupal::service('commerce_iats');
  }

}
