<?php

namespace Drupal\commerce_iats\PluginForm;

use Drupal\commerce_payment\CreditCard;
use Drupal\commerce_payment\Exception\DeclineException;
use Drupal\commerce_payment\PluginForm\PaymentMethodAddForm as BasePaymentMethodAddForm;
use Drupal\Component\Utility\Html;
use Drupal\Core\Form\FormStateInterface;

/**
 * Class PaymentMethodAddForm.
 */
class PaymentMethodAddForm extends BasePaymentMethodAddForm {

  /**
   * {@inheritdoc}
   */
  protected function buildCreditCardForm(array $element, FormStateInterface $form_state) {
    // If not hosted, just do the standard credit card form.
    if (!$this->plugin->isHosted()) {
      return parent::buildCreditCardForm($element, $form_state);
    }

    $config = $this->plugin->getConfiguration();
    $element['#type'] = 'container';
    $element['#attributes']['id'][] = Html::getId('checkout-embed');
    $element['#attributes']['class'][] = Html::getClass('commerce-iats-hosted');
    $element['#attributes']['data-transcenter'] = $config['transcenter'];
    $element['#attributes']['data-processor'] = $config['processor'];
    $element['#attributes']['data-type'] = 'Vault';
    $element['#attached']['library'][] = 'commerce_iats/cryptogram';

    return $element;
  }

  /**
   * {@inheritdoc}
   */
  protected function validateCreditCardForm(array &$element, FormStateInterface $form_state) {
    // Non-hosted, determine credit card type.
    if (!$this->plugin->isHosted()) {
      $values = $form_state->getValue($element['#parents']);
      $card_type = CreditCard::detectType($values['number']);
      if (!$card_type) {
        $form_state->setError($element['number'], t('You have entered a credit card number of an unsupported card type.'));
        return;
      }

      // iATS handles dinersclub as amex.
      $card_type = $card_type->getId();
      if ($card_type == 'dinersclub') {
        $card_type = 'amex';
      }

      // Persist the detected card type.
      $form_state->setValueForElement($element['type'], $card_type);
      return;
    }

    // Handle hosted cryptogram.
    if (!$c = $this->getCryptogram()) {
      throw new DeclineException(t('We encountered an error processing your payment method. Please verify your details and try again.'));
    }

    $key = $element['#parents'];
    array_push($key, 'cryptogram');
    $form_state->setValue($key, $c);
  }

  /**
   * {@inheritdoc}
   */
  public function submitCreditCardForm(array $element, FormStateInterface $form_state) {
    // No op.
  }

  /**
   * Determines if the payment gateway is operating with hosted form processing.
   *
   * @return bool
   *   Indicates if the payment gateway is operating with hosted form
   *   processing.
   */
  protected function isHosted() {
    $config = $this->plugin->getConfiguration();
    return $config['processing_type'] == 'hosted';
  }

  /**
   * Gets the submitted cryptogram value.
   *
   * @return string
   *   The cryptogram value.
   */
  protected function getCryptogram() {
    return $this->getRequestStack()
      ->getCurrentRequest()
      ->request->get('checkout-cryptogram');
  }

  /**
   * Gets the request stack.
   *
   * @return \Symfony\Component\HttpFoundation\RequestStack
   *   The request stack.
   */
  protected function getRequestStack() {
    return \Drupal::requestStack();
  }

}
