<?php

namespace Drupal\commerce_iats\Rest;

/**
 * Interface GatewayInterface.
 */
interface GatewayInterface {

  /**
   * Performs a POST request to the API.
   *
   * @param string $function
   *   The API function to call.
   * @param array $data
   *   Request data, optional.
   *
   * @return array
   *   Response from the request.
   */
  public function post($function, array $data = []);

  /**
   * Credits a previously-settled credit card transaction.
   *
   * @param array $data
   *   Transaction data.
   *
   * @return \stdClass
   *   Credit response data.
   */
  public function creditCardCredit(array $data);

  /**
   * Settles a previously-authorized credit card transaction.
   *
   * @param array $data
   *   Transaction data.
   *
   * @return \stdClass
   *   Settlement response data.
   */
  public function creditCardSettle(array $data);

  /**
   * Voids a previously-authorized credit card transaction.
   *
   * @param array $data
   *   Transaction data.
   *
   * @return \stdClass
   *   Settlement response data.
   */
  public function creditCardVoid(array $data);

  /**
   * Perform authorization with 1stPayVault credit card.
   *
   * @param string $vaultKey
   *   The vault ID.
   * @param string $id
   *   Credit card ID.
   * @param array $data
   *   Credit card data.
   *
   * @return \stdClass
   *   Response from the 1stPayVault credit card authorization.
   */
  public function firstPayCcAuth($vaultKey, $id, array $data);

  /**
   * Perform sale with 1stPayVault credit card.
   *
   * @param string $vaultKey
   *   The vault ID.
   * @param string $id
   *   Credit card ID.
   * @param array $data
   *   Credit card data.
   *
   * @return \stdClass
   *   Response from the 1stPayVault credit card sale.
   */
  public function firstPayCcSale($vaultKey, $id, array $data);

  /**
   * Add a credit card to a vault.
   *
   * @param string $vaultKey
   *   The vault ID.
   * @param array $data
   *   Credit card data.
   *
   * @return \stdClass
   *   Response from the vault add credit card request.
   */
  public function vaultCcCreate($vaultKey, array $data);

  /**
   * Delete a credit card from a vault.
   *
   * @param string $vaultKey
   *   The vault ID.
   * @param string $id
   *   Credit card ID.
   */
  public function vaultCcDelete($vaultKey, $id);

  /**
   * Load credit card data from a vault.
   *
   * @param string $vaultKey
   *   The vault ID.
   * @param string $id
   *   Credit card ID.
   *
   * @return \stdClass|null
   *   Credit card data, or NULL if not found.
   */
  public function vaultCcLoad($vaultKey, $id);

  /**
   * Query credit card data from a vault.
   *
   * @param string $vaultKey
   *   The vault ID.
   * @param array $data
   *   Credit card query data, optional.
   *
   * @return \stdClass[]
   *   Array of credit card data.
   */
  public function vaultCcQuery($vaultKey, array $data = []);

}
